/*
 * keyboard.c
 *
 * Created: 16.02.2020 18:52:49
 *  Author: Chris
 */ 

//             data     add. bytes  comment
// End of Seq  11111111
// Pause       1ttttttt             t >= 0 && t < 127
// TBD         10000000

// Modifier
//   Keep      000xxxxx 
//   Press     001xxxxx +1
//   Hold      010xxxxx +1
//   Release   011xxxxx +1
// Button
//   Keep      0xx00xxx
//   Press     0xx01xxx +1
//   Hold      0xx10xxx +1
//   Release   0xx11xxx +1

#include "keyboard.h"
#include "devicesettings.h"

typedef enum
{
    kbseq_eos_gc         = 0b11111111,
    kbseq_pause_gm       = 0b10000000,
    kbseq_pausetime_gm   = 0b01111111,

    kbseq_mod_gm         = 0b01100000,
    kbseq_mod_keep_gc    = 0b00000000, // just don't modify modifiers
    kbseq_mod_press_gc   = 0b00100000, // press only for the next report
    kbseq_mod_hold_gc    = 0b01100000, // press and hold
    kbseq_mod_release_gc = 0b01000000, // release modifiers (no additional byte is needed)

    kbseq_key_gm         = 0b00011000,
    kbseq_key_keep_gc    = 0b00000000,
    kbseq_key_press_gc   = 0b00001000,
    kbseq_key_hold_gc    = 0b00010000,
    kbseq_key_release_gc = 0b00011000,

}   kbseq_t;

/* = {
    kbseq_mod_press_gc | kbseq_key_press_gc, MOD_SHIFT_LEFT, KEY_H, // H
    kbseq_key_press_gc, KEY_A, // a
    kbseq_key_press_gc, KEY_L, // l
    kbseq_key_press_gc, KEY_L, // l
    kbseq_key_press_gc, KEY_O, // o
    kbseq_pause_gm | 100,
    kbseq_key_press_gc, KEY_SPACE, // 
    kbseq_mod_hold_gc | kbseq_key_press_gc, MOD_SHIFT_RIGHT, KEY_W, // W
    kbseq_key_press_gc, KEY_E, // E
    kbseq_key_press_gc, KEY_L, // L
    kbseq_key_press_gc, KEY_T, // T
    kbseq_mod_release_gc,
    kbseq_key_press_gc, KEY_RETURN,
    kbseq_eos_gc
};*/

kbseqstate_t kbseq;

void keyboard_init(void)
{
    ds_loadkbseq();
    kbseq.running = false;
    kbseq.pause = 0;
    kbseq.blank = false;
}

void keyboard_trigger(void)
{
    if(kbseq.running)
        return;
    kbseq.position = 0;
    kbseq.pause = 0;
    kbseq.running = true;
}

void keyboard_tick()
{
    if(kbseq.running == false)
        return;

    if(kbseq.pause > 0)
    {
        kbseq.pause--;
    }
}

//returns true if a report should be sent
bool keyboard_process(void)
{
    if(kbseq.running == false)
        return false;

    if(kbseq.pause > 0)
    {
        // sequencer is still in pause state, don't do anything (except releasing the key, though: is this really deserved here?) until it's done
        kbseq.keycode = 0;
        return true;
    }

    bool report = false;

    // release buttons they are marked so
    if(kbseq.release_mod == true)
    {
        report = true;
        kbseq.release_mod = false;
        kbseq.modifiers = 0;
    }

    if(kbseq.release_key == true)
    {
        report = true;
        kbseq.release_key = false;
        kbseq.keycode = 0;
    }

    uint8_t cmd = vt_kbseq[kbseq.position];
    kbseq.position++;

    if((cmd & kbseq_pause_gm) != 0)
    {
        if(cmd == kbseq_eos_gc)
        {
            // end of sequence, return
            kbseq.running = false;
            // release the keys
            kbseq.modifiers = 0;
            kbseq.keycode = 0;
            return true;
        }
        
        // if a pause is set, also set the keycode to blank to avoid key repetitions, not quite cool but eh.
        kbseq.pause = (cmd & kbseq_pausetime_gm) + 1;
        kbseq.keycode = 0;
        return true;
    }
    else
    {
        
        uint8_t subcmd = cmd & kbseq_mod_gm;
        if(subcmd != 0)
        {
            report = true;
            if(subcmd == kbseq_mod_release_gc)
            {
                kbseq.modifiers = 0;
            }
            else if((subcmd & kbseq_mod_press_gc) != 0)
            {
                kbseq.modifiers = vt_kbseq[kbseq.position];
                kbseq.position++;

                if((subcmd & kbseq_mod_hold_gc) == kbseq_mod_hold_gc)
                    kbseq.release_mod = false; // release the schmoo! arr, kraken!! err no, just the modifiers.
                else
                    kbseq.release_mod = true;
            }
        }

        subcmd = cmd & kbseq_key_gm;
        if(subcmd != 0)
        {
            report = true;
            if(subcmd == kbseq_key_release_gc)
            {
                kbseq.keycode = 0;
            }
            else if((subcmd & kbseq_key_press_gc) != 0)
            {
                if(kbseq.last_keycode == vt_kbseq[kbseq.position])
                {
                    kbseq.last_keycode = kbseq.keycode; //FIXME

                    // if there's the same keycode, a key release must be inserted
                    // to repeat this key, rewind sequence position
                    kbseq.position--;
                    //if the command was combined with a modifier-command, rewind once more
                    if((cmd & kbseq_mod_gm) != 0)
                        kbseq.position--;

                    kbseq.keycode = 0;
                    return true;
                }
                kbseq.keycode = vt_kbseq[kbseq.position];
                kbseq.last_keycode = kbseq.keycode; //FIXME
                kbseq.position++;

                if((subcmd & kbseq_key_hold_gc) == kbseq_key_hold_gc)
                    kbseq.release_key = false;
                else
                    kbseq.release_key = true;
            }
        }
        

    }
    return report;
}