/* Name: main.c
 * Project: HID-Test
 * Author: Christian Starkjohann
 * Creation Date: 2006-02-02
 * Tabsize: 4
 * Copyright: (c) 2006 by OBJECTIVE DEVELOPMENT Software GmbH
 * License: GNU GPL v2 (see License.txt) or proprietary (CommercialLicense.txt)
 * This Revision: $Id$
 */

// https://forums.obdev.at/viewtopic.php?t=7715


#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <util/delay.h>

#include "usbdrv.h"
#include "oddebug.h"
#include "keyboard.h"
#include "usbdescriptors.h"
#include "regs_tx5.h"
#include "devicesettings.h"

// debouncing minimum ticks (unit: 5 ms) that the button needs to be in particular state 
#define BUTTON_MINTICKS 2

#define ButtonPressed() ((PINB & (1 << PB1)) == 0)

static uchar reportBuffer[8]; //buffer for HID reports
static uchar idleRate; // in 4 ms units

static uchar write_buffer[8]; //buffer for HID write
static uchar write_pos, write_remaining;

usbdescr_t* usbdescr_selected;

void usbEventResetReady()
{
}

uchar usbFunctionDescriptor(usbRequest_t * rq)
{
    uchar len = 0;
    usbMsgPtr = 0;

    if ( rq->wValue.bytes[1] == USBDESCR_STRING && rq->wValue.bytes[0] == 1 ) // Vendor name
    {
        usbMsgPtr = (uchar *) usb_cfg_device_vendorname;
        len = sizeof( usb_cfg_device_vendorname );

    }
    else if(rq->wValue.bytes[1] == USBDESCR_STRING && rq->wValue.bytes[0] == 3) // Serial number
    {
        usbMsgPtr = (uchar *) usb_cfg_device_serialnumber;
        len = sizeof( usb_cfg_device_serialnumber );
    }
    else if(rq->wValue.bytes[1] == USBDESCR_DEVICE )
    {
        usbMsgPtr = (uchar*)usbdescr_selected->device.data;
        len = usbdescr_selected->device.size;
    }
    else if (rq->wValue.bytes[1] == USBDESCR_CONFIG ) 
    {
        usbMsgPtr = (uchar*)usbdescr_selected->config.data;
        len = usbdescr_selected->config.size;
    }
    else if ( rq->wValue.bytes[1] == USBDESCR_STRING && rq->wValue.bytes[0] == 2 ) // device name
    {
        usbMsgPtr = (uchar*)usbdescr_selected->devname.data;
        len = usbdescr_selected->devname.size;
    }
    else if(rq->wValue.bytes[1] == USBDESCR_HID && rq->wValue.bytes[0] == 0)
    {
        usbMsgPtr = (uchar*)usbdescr_selected->hid.data;
        len = usbdescr_selected->hid.size;
    }
    else if(rq->wValue.bytes[1] == USBDESCR_HID_REPORT)
    {
        usbMsgPtr = (uchar*)usbdescr_selected->hidreport.data;
        len = usbdescr_selected->hidreport.size;
    }
    
    return len;
}



uchar usbFunctionSetup(uchar data[8])
{
    usbRequest_t *rq = (void *)data;

    usbMsgPtr = reportBuffer;
    if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS)
    {    /* class request type */
        if(rq->bRequest == USBRQ_HID_GET_REPORT)
        {  
            // FIXME: is key repetition really required?
            if(interface_mode == ifmode_kbd_key)
            {
                //buildReport(keyPressed());
                //return 2;
                return 0;
            }
                return 0;
        }
        else if(rq->bRequest == USBRQ_HID_SET_REPORT)
        {
            if(interface_mode == ifmode_hid)
            {
                write_pos = 0;                // initialize position index
                write_remaining = rq->wLength.word;  // store the amount of data requested
                if(write_remaining > sizeof(write_buffer)) // limit to buffer size
                    write_remaining = sizeof(write_buffer);
                return USB_NO_MSG;        // tell driver to use usbFunctionWrite()
            }
            
        }
        else if(rq->bRequest == USBRQ_HID_GET_IDLE)
        {
            usbMsgPtr = &idleRate;
            return 1;
        }
        else if(rq->bRequest == USBRQ_HID_SET_IDLE)
        {
            idleRate = rq->wValue.bytes[1];
        }
    }
    else
    {
        /* no vendor specific requests implemented */
    }
    return 0;
}

uchar usbFunctionWrite(uchar *data, uchar len)
{
    uchar i;
    if(len > write_remaining)                // if this is the last incomplete chunk
        len = write_remaining;               // limit to the amount we can store
    write_remaining -= len;
    for(i = 0; i < len; i++)
        write_buffer[write_pos++] = data[i];

    if(write_remaining == 0)
    {
        ds_cmdReceived(write_buffer);
    }
    return write_remaining == 0;             // return 1 if we have all data
}

inline void timer1_init()
{
    TCCR1 = (1<<CTC1) | (1<<PWM1A) | (0<<COM1A1) | (0<<COM1A0) | tccr1_clkdiv_256;
    OCR1C = 69; // nice, but no damn clue what's going on there...
}

uint8_t button_debounce = 0;
bool button_value = false;
bool button_changed = false;

#define BUTTON_MASK ((1 << BUTTON_MINTICKS) - 1)

void button_tick(void)
{
    button_debounce <<= 1;
    if(ButtonPressed())
        button_debounce++;

    uint8_t button_masked = button_debounce & BUTTON_MASK;
    bool button_oldstate = button_value;
    if(button_masked == 0)
    {
        button_value = false;
    }
    else if(button_masked == BUTTON_MASK)
    {
        button_value = true;
    }
    if(button_value != button_oldstate)
        button_changed = true;
}

static inline void mainloop_kbd_key()
{
    while (true)
    {
        wdt_reset();
        usbPoll();

        if(TIFR & (1 << TOV1))
        {
            TIFR |= (1 << TOV1);
            button_tick();

            if(button_changed == true)
            {
                if(button_value == true)
                {
                    reportBuffer[0] = vt_key_keycode[0];
                    reportBuffer[1] = vt_key_keycode[1];
                }
                else
                {
                    reportBuffer[0] = 0;
                    reportBuffer[1] = 0;
                }
            }
        }
        
        if(usbInterruptIsReady() && button_changed == true)
        {
            button_changed = false;
            usbSetInterrupt(reportBuffer, 2); // for keyboard, send only 2 bytes
        }
    }
}

static inline void mainloop_kbd_seq(void)
{
    while(true)
    {
        wdt_reset();
        usbPoll();

        if(TIFR & (1<<TOV1))
        {
            TIFR |= (1<<TOV1);

            button_tick();
            keyboard_tick();

            if(button_changed == true && button_value == true)
            {
                button_changed = false;
                keyboard_trigger();
            }
        }

        if(usbInterruptIsReady() && keyboard_process())
        {
            reportBuffer[0] = kbseq.modifiers;
            reportBuffer[1] = kbseq.keycode;
            usbSetInterrupt(reportBuffer, 2); // for keyboard, send only 2 bytes
        }
    }
}

static inline void mainloop_hid(void)
{
    while(true)
    {
        wdt_reset();
        usbPoll();

        if(TIFR & (1<<TOV1))
        {
            TIFR |= (1<<TOV1);
            button_tick();
        }
        ds_task();

        if(usbInterruptIsReady())
        {
            if(button_changed == true)
            {
                button_changed = false;
                reportBuffer[0] = 0; // button state
                reportBuffer[1] = button_value;
                usbSetInterrupt(reportBuffer, sizeof(reportBuffer));
            }
            else if(ds_response.dataready == true)
            {
                usbSetInterrupt(ds_response.data, sizeof(reportBuffer));
                ds_response.dataready = false;
            }
        }
    }
}

static inline void mainloop_pad(void)
{
    while(true)
    {
        wdt_reset();
        usbPoll();

        if(TIFR & (1<<TOV1))
        {
            TIFR |= (1<<TOV1);

            button_tick();
        }

        if(usbInterruptIsReady() && button_changed == true)
        {
            button_changed = false;
            reportBuffer[0] = 0;
            if(button_value)
            {
                reportBuffer[0] = 1;
            }
            usbSetInterrupt(reportBuffer, 1);
        }
    }
}

int main(void)
{
    wdt_enable(WDTO_2S);
    
    DDRB &= ~(1 << PB1);
    _delay_ms(10); //let the IO settle

    ds_loadifmode();
    
    // when the mode is invalid or the button is pressed at startup,
    // go to HID mode (which is also programming mode)
    if(ButtonPressed())
        interface_mode = ifmode_hid;

    switch (interface_mode)
    {
        case ifmode_pad: 
            usbdescr_selected = &usbdescr_pad; 
            break;
        case ifmode_kbd_key:
        case ifmode_kbd_seq:
            usbdescr_selected = &usbdescr_kb;
            break;
        case ifmode_hid:
        default:
            usbdescr_selected = &usbdescr_hid; 
            break;
    }

    odDebugInit();
    usbInit();
    sei();
    DBG1(0x00, 0, 0);
    timer1_init();

    if(interface_mode == ifmode_kbd_seq)
    {
        keyboard_init();
        mainloop_kbd_seq();
    }
    else if(interface_mode == ifmode_kbd_key)
    {
        ds_loadkbdkey();
        mainloop_kbd_key();

    }
    else if(interface_mode == ifmode_pad)
    {
        mainloop_pad();
    }
    else
    {
        // for all other (and maybe invalid?) modes, go to hid mode
        mainloop_hid();
    }

    return 0;
}
