# -*- coding: utf-8 -*-

import hid
import math
from enum import Enum

class FootswitchException(Exception):
    pass

class Footswitch():

    class Command(Enum):
        nop = 0
        restart = 1
        get_ifmode = 2
        set_ifmode = 3
        get_keyboard_key = 4
        set_keyboard_key = 5
        get_keyboard_seq = 6
        set_keyboard_seq = 7
    
    class IfMode(Enum):
        hid = 0
        kbd_key = 1
        kbd_seq = 2
        pad = 3

    def __init__(self):
        devices = hid.enumerate(0x16C0, 0x27D9)
        if len(devices) == 0:
            raise FootswitchException("no devices found")

        self.dev = hid.device()
        self.dev.open_path(devices[0]['path'])

    def restart(self):
        cmd = self.Command.restart
        self.dev.write([0, cmd.value, 0, 0, 0, 0, 0, 0, 0])
    
    def sendcommand(self, cmd, data = None):
        if not isinstance(cmd, self.Command):
            raise FootswitchException('Wrong datatype for cmd')

        out = [0, cmd.value]
        if data is None:
            data = [0]
        data += [0] * 7 # add some padding
        out += data[0:7] # and add truncated to out data
        result = self.dev.write(out)
        if result == -1:
            raise FootswitchException('Could not write data to device')
        
        result = self.dev.read(8)
        
        while result[0] == 0:
            # TESTME: as long as the first byte of the report is 0,
            # button status reports are received. Read until another report is received
            result = self.dev.read(8)
        
        if result[0] & 0x3F is not cmd.value:
            raise FootswitchException('Response is not for requested data: expected 0x%02X, got 0x%02X' % (cmd.value, result[0] & 0x3F))
        if result[0] >> 6 is not 2:
            raise FootswitchException('Response is not ok (%i)' % (result[0] >> 6))
        return result[1:]

    def get_interface_mode(self):
        data = self.sendcommand(self.Command.get_ifmode)
        return self.IfMode(data[0])

    def set_interface_mode(self, mode):
        if not isinstance(mode, self.IfMode):
            raise FootswitchException('Wrong datatype for cmd, use Footswitch.IfMode')
        self.sendcommand(self.Command.set_ifmode, [mode.value])

    def get_kbmode_key(self):
        data = self.sendcommand(self.Command.get_keyboard_key)
        return data[0:2]

    def set_kbmode_key(self, modifier, keycode):
        self.sendcommand(self.Command.set_keyboard_key, [modifier, keycode])
    
    def set_kbseq(self, sequence):
        datalen = 128
        chunklen = 6

        if len(sequence) > datalen:
            raise FootswitchException('sequence too long for available space')

        chunks = math.ceil(len(sequence) / chunklen)

        for c in range(0, chunks):
            data = [c] + sequence[c * chunklen : (c + 1) * chunklen] # chunk number + data
            data += [255] * (chunklen - len(data) + 1)
            #print("writing data", data)
            self.sendcommand(self.Command.set_keyboard_seq, data)
    
    def get_kbseq(self):
        datalen = 128
        chunklen = 6

        chunk = 0
        data = []
        while datalen > 0:
            result = self.sendcommand(self.Command.get_keyboard_seq, [chunk])
            copylen = min(chunklen, datalen)
            #print("copy chunk %u [%u:%u] ->" % (chunk, 1, 1 + copylen), result[1 : 1 + copylen])
            data += result[1:1+copylen]
            chunk += 1
            datalen -= 6
        #print("all data:", ' '.join('{:02x}'.format(x) for x in data))
        #print(data)
        return data

    def hid_waitbutton(self):
        while True:
            # use a loop to ignore every other communication
            data = self.dev.read(8)
            if data[0] == 0:
                if data[1] == 0:
                    return False
                elif data[1] == 1:
                    return True
                else:
                    return None

class FootswitchSequence():

    class FootswitchCommands(Enum):
        eos_gc         = 0xFF
        pause_gm       = 0x80
        pausetime_gm   = 0x7F
        mod_gm         = 0x60
        mod_keep_gc    = 0x00
        mod_press_gc   = 0x20
        mod_hold_gc    = 0x60
        mod_release_gc = 0x40
        key_gm         = 0x18
        key_keep_gc    = 0x00
        key_press_gc   = 0x08
        key_hold_gc    = 0x10
        key_release_gc = 0x18

    @classmethod
    def keycodes_to_commands(cls, codes):
        cmds = []
        for c in codes:
            cmd = cls.FootswitchCommands.key_press_gc.value
            key = c[1]
            mod = c[2]
            if mod != 0:
                cmd |= cls.FootswitchCommands.mod_press_gc.value
            
            cmds.append(cmd)

            if mod != 0:
                cmds.append(mod)
            
            cmds.append(key)
        
        cmds.append(cls.FootswitchCommands.eos_gc.value)
        return cmds